#!/usr/bin/env python
import os
import errno
import boto3
from botocore.exceptions import EndpointConnectionError,ClientError  # type: ignore
import logging
from datetime import datetime, timezone
from dateutil import parser
import s3_bk_files.config as config


class AwsS3Helper:
    
    def __init__(self,AWS_ACCESS_KEY_ID,AWS_SECRET_ACCESS_KEY,region_name="us-east-1"):
        '''
             Method to programatically access S3 objects.
             PARAMETERS:
                AWS_ACCESS_KEY_ID (string): Access Key associated with an AWS account.
                AWS_SECRET_ACCESS_KEY (string): Corresponding Secret Key associated to the account.
                (OPTIONAL) region_name (string): Region for the account. Set default to 'us-east-1'. 
        '''
     
        self.conn = boto3.client("s3", region_name = region_name, aws_access_key_id = AWS_ACCESS_KEY_ID, \
                    aws_secret_access_key = AWS_SECRET_ACCESS_KEY)  
            
         
        
        
      
    def set_key(self, path):
        '''
            Utility function to get bucket and key from a S3 path
            PARAMETERS:
                path (string): s3 path
            OUTPUT:
                bucket (string): bucket in which the s3 path is
                key (string): s3 path without the bucket.
            USAGE:
                set_key('s3_bucket/s3_path/file.txt')
        '''
        bucket = path.split("/")[0]
        key = "/".join(path.split("/")[1:])
        return bucket, key

    def upload_file(self, path, inpfname):
        '''
            To upload a S3 object
            PARAMETERS:
                path (string): s3 path where file to be uploaded. Should be specified as bucket + key.
                inpfname (string): local path of file to be uploaded
            RETURNS: None
            USAGE:
                upload_file('s3_bucket/s3_path/file.txt','/home/abc/local_dir/file.txt')
        '''
        bucket, key = self.set_key(path)
        self.conn.upload_file(inpfname, bucket, key)
    
    def download_file(self, path, outfname):
        '''
            To download a S3 object.
            PARAMETERS:
                path (string): s3 object to be downloaded. Should be specified as bucket + key.
                inpfname (string): local path of where the object is to be downloaded.
            RETURNS: None
            USAGE:
                download_file('s3_bucket/s3_path/file.txt','/home/abc/local_dir/file.txt')
        '''
        try:
            bucket, key = self.set_key(path)
            self.conn.download_file(bucket, key, outfname)
        except ClientError as e:
            logging.error(e)
        except OSError  as e:
            if e.errno == errno.ENOSPC:
                logging.error(' No hay espacio suficiente en el disco.')
            else:
                logging.error(e)
        except Exception as e:
            print('Exeption: ',e)

    def download_file_from_dict(self, file_dict, outfname):
        path = file_dict['S3path']
        return self.download_file(path, outfname)
        

    def del_file(self, path):
        '''
            To delete a s3 object.
            PARAMETERS: 
                path (string): s3 path of object to be deleted.
            RETURNS: None
            USAGE:
                del_file('s3_bucket/s3_path/file.txt')
        '''
        bucket, key = self.set_key(path)
        self.conn.delete_object(Bucket=bucket,Key=key)

    def list_keys(self, prefix_name):
        ''' 
            Used to list files in a S3 object/folder.
            PARAMETERS:
                prefix_name (string): s3 path of folder/object
            OUPTUT:
                A dict containing s3 paths in "S3path" key .
            USAGE:
                list_keys('s3_bucket/s3_path/')
        '''
        bucket, key = self.set_key(prefix_name)
        is_truncated = True
        result = []
        marker = ""
        while is_truncated:
            response = self.conn.list_objects(Bucket=bucket,Prefix=key,Delimiter=",",Marker=marker)
            if "Contents" not in response: break
            result.extend(response["Contents"])
            is_truncated = response["IsTruncated"]
            if "NextMarker" not in response: break
            marker = response["NextMarker"]
        return map(lambda x:{"S3path":bucket + "/" + x["Key"],"info":x}, result)

    def sorted_list_keys(self, prefix_name, reverse=False):
        ''' 
            PARAMETERS:
                prefix_name (string): s3 path of folder/object
            OUPTUT:
                A dict containing s3 paths in "S3path" key .
            USAGE:
                sorted_list_keys('s3_bucket/s3_path/')
        '''
        map_list = list(self.list_keys(prefix_name))
        # Change UTC ZONE TO LOCAL
        if map_list:
            utc_map_list = []
            for obj in map_list :
                obj['info']['LastModified'] = self.utc_to_local(obj['info']['LastModified'])
                utc_map_list.append(obj)
            map_list = utc_map_list
            get_last_modified = lambda obj: int(obj['info']['LastModified'].strftime('%s'))
            return sorted(map_list, key=get_last_modified, reverse=reverse)
        else:
            raise Exception("No hay archivos a procesar en el bucket. Por favor, revise el archivo de configuración o confirme que no hay archivos en el Bucket configurado")

    def utc_to_local(self,utc_dt):
        return utc_dt.replace(tzinfo=timezone.utc).astimezone(tz=None)

    def aslocaltimestr(self,utc_dt):
        return self.utc_to_local(utc_dt).strftime(config.datetime_format_string)

    def filter_by_datetime(self, list_keys, last_datetime):
        ''' 
            PARAMETERS:
                prefix_name (string): s3 path of folder/object
            OUPTUT:
                A dict containing s3 paths in "S3path" key .
            USAGE:
                Retorna una lista con los archivos posteriores
                a la fecha indicada
        '''
        filter_list = []
        if last_datetime:
            last_datetime_dt = parser.parse(last_datetime)
            
            for obj in list_keys:
                if obj['info']['LastModified'].replace(tzinfo=None) > last_datetime_dt:
                    filter_list.append(obj)

        return filter_list

                
    def copy_key(self, src_key, dst_key):
        '''
            Used to copy file from one s3 path to another
            PARAMETERS:
                src_key (string): s3 path from where file to be copied
                dst_key (string): s3 path where file to be copied
            OUTPUT: None
            USAGE:
                copy_key('s3_bucket1/some_path/fname.txt','s3_bucket2/some_path2/fname2.txt')
        '''
        bucket ,key = self.set_key(dst_key)
        self.conn.copy_object(Bucket=bucket,CopySource=src_key,Key=key)

    def get_last_modified(self, path):
        '''
            To get last modified timestamp of an object
            PARAMETERS:
                path (string): s3 path of object
            RETURNS: Python datetime object
            USAGE:
                get_last_modified('s3_bucket/s3_path/file.txt')
        '''
        bucket, key = self.set_key(path)
        object_info = self.resource.Object(bucket, key)
        timestamp = object_info.last_modified
        return timestamp

    def clear_s3_folder(self,s3_path):
        '''
            Utility function to delete a s3 path and it's subpaths.
            PARAMETERS:
                s3_path (string): s3 path to be cleared.
            OUPTUT: None
            USAGE:
                clear_s3_folder('s3_bucket/some_path/')
        '''
        file_keys = self.conn_s3.list_keys(s3_path)
        for fkey in file_keys:
            self.conn_s3.del_file(fkey)
